package org.agil.core.spatialindex;

import java.util.Stack;

/**
 *  Implementa las funciones basicas del interface Node, asi como las
 *  funciones de splitting. <p>
 *
 *  Created: Tue May 18 15:57:56 1999 <p>
 *
 *
 *
 *@author     Marios Hadjieleftheriou
 *@version    2
 */
public abstract class AbstractNode implements Node {
	/**
	 *  Nivel de este nodo. Los nodos de tipos Leaf siempre tienen un nivel igual
	 *  a 0.
	 */
	protected int level;
	/**
	 *  Padre de todos los nodos
	 */
	protected transient RTree tree;
	/**
	 *  El pageNumber donde el padre de este nodo esta almacenado.
	 */
	protected long parent;
	/**
	 *  El pageNumber donde este nodo esta almacenado.
	 */
	protected long pageNumber;
	/**
	 *  Todos los nodos estan almacenados en este Array. Este debe tener un
	 *  tamano de <b>nodeCapacity +1 </b> para alojar todos los datos mas un
	 *  HyperCube, cuando el nodo debe ser dividido.
	 */
	protected HyperCube[] data;
	/**
	 *  Contiene los pageNumbers que apuntan a los hijos de este nodo. Siempre
	 *  tiene el mismo tamano que el Array <i>data</i> . Si este es un nodo de
	 *  tipo Leaf, todos los branches deberian apuntar a objetos reales.
	 */
	protected long[] branches;
	/**
	 *  Cuanto espacio esta usado en este nodo. Si es igual a nodeCapacity
	 *  entonces el nodo esta lleno
	 */
	protected int usedSpace;
	/**
	 *  El minimo MBB(=MER) de los datos contenidos en este nodo.
	 */
	protected HyperCube nodeMbb;


	/**
	 *  Crea un AbstracNode a partir de los parametros:
	 *
	 *@param  tree        RTree al que pertenece este nodo.
	 *@param  parent      nodo padre de este nodo.
	 *@param  pageNumber  numero de pagina donde esta almacenado este nodo.
	 *@param  level       nivel del RTree donde esta almacenado
	 */
	protected AbstractNode(RTree tree, long parent, long pageNumber, int level) {
		this.parent = parent;
		this.tree = tree;
		this.pageNumber = pageNumber;
		this.level = level;
		data = new HyperCube[tree.getNodeCapacity() + 1];
		branches = new long[tree.getNodeCapacity() + 1];
		usedSpace = 0;
		nodeMbb = null;
	}


	//
	// Implementacion del Interface Node.java
	//

	/**
	 *  Retorna el nivel del nodo. Siempre cero si es una hoja (=Leaf).
	 *
	 *@return    Nivel del nodo.
	 */
	public int getLevel() {
		return level;
	}


	/**
	 *@return    true si este nodo es el nodo Root.
	 */
	public boolean isRoot() {
		return (parent == RTree.NIL);
	}


	/**
	 *@return    true si este nodo es de tipo Index. El Root tambien es de tipo
	 *      Index, a menos que sea una hoja.
	 */
	public boolean isIndex() {
		return (level != 0);
	}


	/**
	 *  return true si este nodo es de tipo Leaf. El Root tambien puede ser de
	 *  tipo Leaf.
	 *
	 *@return    The Leaf value
	 */
	public boolean isLeaf() {
		return (level == 0);
	}


	/**
	 *  Retorna el MBB de todos los datos contenidos en este nodo.
	 *
	 *@return    El MBB de los datos.
	 */
	public HyperCube getNodeMbb() {
		if (nodeMbb == null) {
			return null;
		}
		else {
			return (HyperCube) nodeMbb.clone();
		}
	}


	/**
	 *  Retorna una copia de los HyperCubes presentes en este nodo.
	 *
	 *@return    un array de HyperCubes conteniendo copias de los datos
	 *      originales.
	 */
	public HyperCube[] getHyperCubes() {
		HyperCube[] h = new HyperCube[usedSpace];
		for (int i = 0; i < usedSpace; i++) {
			h[i] = (HyperCube) data[i].clone();
		}
		return h;
	}


	/**
	 *@return    los branches contenidos por este nodo
	 */
	public long[] getBranches() {
		return branches;
	}


	/**
	 *  Implementacion de Object.toString()
	 *
	 *@return    A string representation of the object.
	 */
	public String toString() {
		String s = "< Page: " + pageNumber + ", Level: " + level + ", UsedSpace: " + usedSpace + ", Parent: " + parent + "\n";
		for (int i = 0; i < usedSpace; i++) {
			s += "  " + (i + 1) + ") " + data[i].toString() + " --> " + " points to: " + branches[i] + "\n";
		}
		s += ">\n";
		return s;
	}


	/**
	 *  Retorna el padre de este nodo. SI tiene padre, debe ser un Index. Si este
	 *  nodo es el Root, retorna null. Esta funcion carga una pagina de datos en
	 *  la memoria principal.
	 *
	 *@return    el nodo padre de este nodo.
	 */
	protected AbstractNode getParent() {
		if (isRoot()) {
			return null;
		}
		else {
			return tree.file.readNode(parent);
		}
	}


	//
	// Metodos utiles.
	//

	/**
	 *  Actualiza el MBB de todos los HyperCubes contenidos en este nodo.
	 */
	protected void updateNodeMbb() {
		if (usedSpace > 0) {
			nodeMbb = (HyperCube) data[0].clone();
			for (int i = 1; i < usedSpace; i++) {
				nodeMbb = nodeMbb.combinedMbb(data[i]);
			}
		}
		else {
			// return an empty mbb.
			nodeMbb = null;
		}
	}


	//
	// Metodos abstractos.
	//

	/**
	 *  Encuentra las hojas mas apropiadas donde el HyperCube dado deberia ser
	 *  almacenado.
	 *
	 *@param  h  El nuevo HyperCube.
	 *@return    la hoja donde el nuevo HyperCube deberia ser insertado.
	 */
	protected abstract Leaf chooseLeaf(HyperCube h);


	/**
	 *  retorna la hoja que contiene el HyperCube dado, o null si el HyperCube no
	 *  esta contenido en ninguna de las hojas de este nodo.
	 *
	 *@param  h  Description of Parameter
	 *@return    Description of the Returned Value
	 */
	protected abstract Leaf findLeaf(HyperCube h);


	/**
	 *  Description of the Method
	 *
	 *@param  h    Description of Parameter
	 *@param  pos  Description of Parameter
	 *@return      Description of the Returned Value
	 */
	protected abstract Leaf findLeaf(HyperCube h, long pos);


	//
	// Metodos de insercion/borrado.
	//

	/**
	 *  Anade un nodo hijo a este nodo. Esta funcion no salva el nodo en el
	 *  disco. Es usado para la carga de un nodo con datos. El usuario debe
	 *  asegurarse de salvar el nodo en disco despues de llamar a esta funcion.
	 *
	 *@param  n  el nuevo nodo a insertar como hijo del nodo actual.
	 */
	protected void addData(AbstractNode n) {
		addData(n.nodeMbb, n.pageNumber);
	}


	/**
	 *  Anade un nodo hijo a este nodo. Esta funcion no salva el nodo en el
	 *  disco. Es usado para la carga de un nodo con datos. El usuario debe
	 *  asegurarse de salvar el nodo en disco despues de llamar a esta funcion.
	 *
	 *@param  h     The feature to be added to the Data attribute
	 *@param  page  la pagina donde este nodo esta localizado, si lo esta.
	 */
	protected void addData(HyperCube h, long page) {
		/*
		 *  if (usedSpace == tree.getNodeCapacity()) {
		 *  throw new IllegalStateException("Node is full.");
		 *  }
		 */
		data[usedSpace] = (HyperCube) h.clone();
		branches[usedSpace] = page;
		usedSpace++;
		if (nodeMbb != null) {
			nodeMbb = nodeMbb.combinedMbb(h);
		}
		else {
			nodeMbb = (HyperCube) h.clone();
		}
	}


	/**
	 *  Borra los datos de una entrada de este nodo. Esta funcion no salva los
	 *  datos en disco. El usuario debe asegurarse de salvar el nodo en disco
	 *  despues de llamar a esta funcion.
	 *
	 *@param  i  el indice de los datos a ser borrados.
	 */
	protected void deleteData(int i) {
		if (usedSpace == 1) {
			// only one element is a special case.
			data[0] = null;
			branches[0] = 0;
		}
		else {
			System.arraycopy(data, i + 1, data, i, usedSpace - i - 1);
			System.arraycopy(branches, i + 1, branches, i, usedSpace - i - 1);
		}
		usedSpace--;
		updateNodeMbb();
	}


	/**
	 *  Algoritmo de Guttman para dividir un nodo. [A. Guttman 'R-trees a dynamic
	 *  index structure for spatial searching']
	 *
	 *@param  h     el HyperCube que causa la division.
	 *@param  page  la pagina donde el nodo hijo que causa la division esta
	 *      almacenado. El id del objeto de los nuevos datos si la division
	 *      ocurre en una hoja.
	 *@return       lso dos grupos de datos que deberian ser asignado a cada
	 *      nuevo nodo.
	 */
	protected int[][] rtreeSplit(HyperCube h, long page) {
		/*
		 *  if (h == null) {
		 *  throw new IllegalArgumentException("Hypercube cannot be null.");
		 *  }
		 */
		// inserta temporalmente un nuevo hypercube en data.
		// El array de datos es siempre uno mas que la capacidad del nodo.
		data[usedSpace] = h;
		branches[usedSpace] = page;
		int total = usedSpace + 1;
		// usa este array para marcar las entradas visitadas(inicializado a false por defecto):
		boolean[] mask = new boolean[total];
		// el numero minimo de entradas que un nodo debe contener
		int min = tree.getMinimumLoad();
		// cuantos entradas quedan si verificar
		int rem = total;
		// crea dos grupos de entradas, para la division del nodo en dos.
		int[] g1 = new int[tree.getNodeCapacity()];
		int[] g2 = new int[tree.getNodeCapacity()];
		// mantiene el ultimo elemento insertado de cada grupo
		int i1 = 0;
		int i2 = 0;
		// inicializa cada grupo
		int[] seed = pickSeeds();
		g1[i1++] = seed[0];
		g2[i2++] = seed[1];
		rem -= 2;
		mask[g1[0]] = true;
		mask[g2[0]] = true;
		// busca los mbr de cada uno de los grupos
		HyperCube mbr1 = (HyperCube) data[g1[0]].clone();
		HyperCube mbr2 = (HyperCube) data[g2[0]].clone();
		while (rem > 0) {
			if (min - i1 == rem) {
				// asignamos entradas al grupo1
				for (int i = 0; i < total; i++) {
					if (mask[i] == false) {
						g1[i1++] = i;
						mask[i] = true;
						rem--;
					}
				}
			}
			else if (min - i2 == rem) {
				// asignamos entradas al grupo2
				for (int i = 0; i < total; i++) {
					if (mask[i] == false) {
						g2[i2++] = i;
						mask[i] = true;
						rem--;
					}
				}
			}
			else {
				double md1 = 0;
				double md2 = 0;
				int sel = -1;
				// para cada entrada que todavia no ha sido asignada a un grupo, determinamos el coste
				// de ponerlo en uno u otro grupo, y seleccionamos aquel con las maxima diferencia entre
				// los costes. En caso de un Linear Split, seleccionamos cualquiera (p.ej. el primero)
				double max = Double.NEGATIVE_INFINITY;
				double d1;
				double d2;
				double d;
				for (int i = 0; i < total; i++) {
					if (mask[i] == false) {
						HyperCube a = mbr1.combinedMbb(data[i]);
						d1 = a.getArea() - mbr1.getArea();
						HyperCube b = mbr2.combinedMbb(data[i]);
						d2 = b.getArea() - mbr2.getArea();
						d = Math.abs(d1 - d2);
						if (d > max) {
							max = d;
							md1 = d1;
							md2 = d2;
							sel = i;
							if (tree.getTreeType() == RTree.RTREE_LINEAR) {
								break;
							}
						}
					}
				}
				// determinamos el grupo donde deberia anadirse la nueva entrada
				int group = -1;
				if (md1 < md2) {
					g1[i1++] = sel;
					group = 1;
				}
				else if (md2 < md1) {
					g2[i2++] = sel;
					group = 2;
				}
				else if (mbr1.getArea() < mbr2.getArea()) {
					g1[i1++] = sel;
					group = 1;
				}
				else if (mbr2.getArea() < mbr1.getArea()) {
					g2[i2++] = sel;
					group = 2;
				}
				else if (i1 < i2) {
					g1[i1++] = sel;
					group = 1;
				}
				else if (i2 < i1) {
					g2[i2++] = sel;
					group = 2;
				}
				else {
					g1[i1++] = sel;
					group = 1;
				}
				mask[sel] = true;
				rem--;
				if (group == 1) {
					mbr1 = mbr1.combinedMbb(data[sel]);
				}
				else {
					mbr2 = mbr2.combinedMbb(data[sel]);
				}
			}
		}
		// retornamos los dos grupos. Dejamos que la subclase decida que hacer con ellos.
		int[][] ret = new int[2][];
		ret[0] = new int[i1];
		ret[1] = new int[i2];
		System.arraycopy(g1, 0, ret[0], 0, i1);
		System.arraycopy(g2, 0, ret[1], 0, i2);
		return ret;
	}


	/*
	 *  protected Node[] rstarSplit(HyperCube h, int page) {
	 *  // chooseSplitAxis.
	 *  for (int i = 0; i < tree.getDimension(); i++) {
	 *  HyperCube[] data = getData();
	 *  final int dim = i;
	 *  Sort.mergeSort(
	 *  data,
	 *  new Comparator() {
	 *  public int compare(Object o1, Object o2) {
	 *  float ret = ((HyperCube) o1).getP1().getFloatCoordinate(dim) -
	 *  ((HyperCube) o2).getP1().getFloatCoordinate(dim);
	 *  if (ret > 0) return 1;
	 *  else if (ret < 0) return -1;
	 *  else  return 0;
	 *  }
	 *  }
	 *  );
	 *  for (int j = 0; j < data.length; j++)
	 *  System.out.println(data[j]);
	 *  }
	 *  return null;
	 *  }
	 */
	/**
	 *  Esta funcion es usada al dividir para inicializar los dos grupos de
	 *  HyperCubes. [A. Guttman 'R-trees a dynamic index structure for spatial
	 *  searching'] 3.5.2. A Quadratic-Cost Algorithm 3.5.3. A Linear-Cost
	 *  Algorithm
	 *
	 *@return    los dos indices de las entradas seleccionadas como primeros
	 *      elementos de cada grupo
	 */
	protected int[] pickSeeds() {
		int i1 = -1;
		int i2 = -1;
		switch (tree.getTreeType()) {
			case RTree.RTREE_LINEAR:
				float separation = Float.NEGATIVE_INFINITY;
				for (int dim = 0; dim < tree.getDimension(); dim++) {
					float leastLower = data[0].getLowPoint().getFloatCoordinate(dim);
					float greatestUpper = data[0].getHighPoint().getFloatCoordinate(dim);
					int greatestLower = 0;
					int leastUpper = 0;
					float width;
					for (int i = 1; i <= usedSpace; i++) {
						if (data[i].getLowPoint().getFloatCoordinate(dim) > data[greatestLower]
								.getLowPoint().getFloatCoordinate(dim)) {
							greatestLower = i;
						}
						if (data[i].getHighPoint().getFloatCoordinate(dim) < data[leastUpper]
								.getHighPoint().getFloatCoordinate(dim)) {
							leastUpper = i;
						}
						leastLower = Math.min(data[i].getLowPoint().getFloatCoordinate(dim), leastLower);
						greatestUpper = Math.max(data[i].getHighPoint().getFloatCoordinate(dim), greatestUpper);
					}
					width = greatestUpper - leastLower;
					width = (width > 0) ? width : 1;
					float f = (data[greatestLower].getLowPoint().getFloatCoordinate(dim) - data[leastUpper]
							.getHighPoint().getFloatCoordinate(dim)) / width;
					if (f > separation) {
						i1 = leastUpper;
						i2 = greatestLower;
						separation = f;
					}
				}
				break;
			case RTree.RTREE_QUADRATIC:
				double inefficiency = Double.NEGATIVE_INFINITY;
				// for each pair of HyperCubes (account for overflow HyperCube too!)
				for (int i = 0; i < usedSpace; i++) {
					for (int j = i + 1; j <= usedSpace; j++) {
						// get the mbr of those two entries.
						HyperCube h = data[i].combinedMbb(data[j]);
						// find the inefficiency of grouping these entries together.
						double d = h.getArea() - data[i].getArea() - data[j].getArea();
						if (d > inefficiency) {
							inefficiency = d;
							i1 = i;
							i2 = j;
						}
					}
				}
				break;
			default:
				throw new IllegalStateException("Linear or Quadratic tree type needed.");
		}
		return new int[]{i1, i2};
	}


	/**
	 *  Utilizado para reordenar el arbol despues de que una entrada haya sido
	 *  borrada [A. Guttman 'R-trees a dynamic index structure for spatial
	 *  searching'] 3.3. Deletion
	 *
	 *@param  s  Description of Parameter
	 */
	protected void condenseTree(Stack s) {
		if (isRoot()) {
			// eliminate root if it has only one child.
			if (!isLeaf() && usedSpace == 1) {
				AbstractNode n = tree.file.readNode(branches[0]);
				tree.file.deletePage(n.pageNumber);
				n.pageNumber = 0;
				n.parent = RTree.NIL;
				tree.file.writeNode(n);
				if (!n.isLeaf()) {
					for (int i = 0; i < n.usedSpace; i++) {
						AbstractNode m = ((Index) n).getChild(i);
						m.parent = 0;
						tree.file.writeNode(m);
					}
				}
			}
		}
		else {
			// find the parent.
			AbstractNode p = getParent();
			int e = -1;
			// find the entry in the parent, that points to this node.
			for (e = 0; e < p.usedSpace; e++) {
				if (pageNumber == p.branches[e]) {
					break;
				}
			}
			int min = tree.getMinimumLoad();
			if (usedSpace < min) {
				// eleminate node.
				p.deleteData(e);
				s.add(this);
			}
			else {
				// adjust the entry in 'p' to contain the new bounding box of this node.
				p.data[e] = getNodeMbb();
				p.updateNodeMbb();
			}
			tree.file.writeNode(p);
			p.condenseTree(s);
		}
	}
}
// AbstractNode
