package org.agil.core.spatialindex;

import java.util.*;

/*
 *  A cached persistent page file with a Least Recently Used (LRU), write back or write through caching policy.<br>
 *  <b>Remeber to flush the cache before exiting, if write back policy is used!</b>.
 *  <p>
 *  Created: Sun Oct 31 19:22:05 1999
 *  <p>
 *  @author Marios Hadjieleftheriou
 *  @version 2
 */

/**
 *  PersistentPageFile con una cache que utiliza Least Recently Used (LRU)
 *  para determinar que elementos son descartados o insertados en la cache.
 *
 *@author    Chris Seguin
 */
public class CachedPersistentPageFile extends PersistentPageFile {
	private HashMap cache;
	private LinkedList lru;
	/**
	 *  Capacidad maxima de nodos.
	 */
	private int capacity = 0;
	/**
	 *  Politica de Escritura. Write back or write through.
	 */
	private boolean writeBack = true;


	/**
	 *  Construye un CachedPersistentPageFile. La capacidad de la cache es 10
	 *  nodos.
	 *
	 *@param  fileName  Nombre del fichero
	 */
	public CachedPersistentPageFile(String fileName) {
		this(fileName, 10, true);
	}


	/**
	 *  Construye un CachedPersistentPageFile.
	 *
	 *@param  fileName  Nombre del fichero
	 *@param  capacity  Capacidad de la cache
	 */
	public CachedPersistentPageFile(String fileName, int capacity) {
		this(fileName, capacity, true);
	}


	/**
	 *  Construye un CachedPersistentPageFile.
	 *
	 *@param  fileName   Nombre del fichero
	 *@param  capacity   Capacidad de la cache
	 *@param  writeBack  Description of Parameter
	 */
	public CachedPersistentPageFile(String fileName, int capacity, boolean writeBack) {
		super(fileName);
		if (capacity < 1) {
			throw new IllegalArgumentException();
		}
		cache = new HashMap(capacity);
		lru = new LinkedList();
		this.capacity = capacity;
		this.writeBack = writeBack;
	}


	/**
	 *  Creates an exact copy of this object.
	 *
	 *@return    Description of the Returned Value
	 */
	public Object clone() {
		return this;
	}


	/**
	 *  Si la politica de escritura es writeBack, escribe los nodos de la cache
	 *  que han sido modificados en el fichero.
	 */
	public void flush() {
		if (writeBack) {
			for (int i = 0; i < lru.size(); i++) {
				CacheEntry e = (CacheEntry) cache.get(lru.get(i));
				if (e.dirty) {
					super.writeNode(e.node);
				}
			}
		}
	}


	/**
	 *  Limpia la cache, escribiendo los nodos modificados.
	 */
	public void clearCache() {
		if (writeBack) {
			for (int i = 0; i < lru.size(); i++) {
				CacheEntry e = (CacheEntry) cache.get(lru.get(i));
				if (e.dirty) {
					super.writeNode(e.node);
				}
			}
		}
		cache.clear();
		lru.clear();
	}


	/**
	 *  Finalizador. Hace flush del fichero.
	 *
	 *@exception  Throwable  Description of Exception
	 */
	protected void finalize() throws Throwable {
		flush();
		super.finalize();
	}


	/**
	 *  Lee una pagina del fichero o de la cache.
	 *
	 *@param  page                pagina a leer
	 *@return                     Description of the Returned Value
	 *@exception  PageFaultError  Description of Exception
	 */
	protected AbstractNode readNode(long page) throws PageFaultError {
		return get(new Long(page));
	}


	/**
	 *  Escribe un nodo en la cache y en el fichero.
	 *
	 *@param  n                   nodo a escribir
	 *@return                     Description of the Returned Value
	 *@exception  PageFaultError  Description of Exception
	 */
	protected long writeNode(AbstractNode n) throws PageFaultError {
		long page = n.pageNumber;
		if (!writeBack || page <= 0) {
			page = super.writeNode(n);
		}
		put(new Long(page), n);
		return page;
	}


	/**
	 *  Borra una pagina del fichero.
	 *
	 *@param  page                pagina a borrar
	 *@return                     objeto borrado
	 *@exception  PageFaultError  Description of Exception
	 */
	protected AbstractNode deletePage(long page) throws PageFaultError {
		Long key = new Long(page);
		if (cache.containsKey(key)) {
			lru.remove(key);
			cache.remove(key);
		}
		return super.deletePage(page);
	}


	/**
	 *  Obtiene una pagina. Si dicha pagina esta previamente en la cache, no va a
	 *  buscarla a disco
	 *
	 *@param  key  Description of Parameter
	 *@return      Description of the Returned Value
	 */
	private AbstractNode get(Long key) {
		if (cache.containsKey(key)) {
			// remove from LRU queue and add it at the end.
			lru.remove(key);
			lru.add(key);
			tree.stats.hits++;
			return ((CacheEntry) cache.get(key)).node;
		}
		else {
			AbstractNode n = super.readNode(key.longValue());
			add(key, new CacheEntry(n));
			tree.stats.misses++;
			return n;
		}
	}


	/**
	 *  Comprueba si tiene la key en la cache, si es asi, la borra de su posicion
	 *  actual en la lista y la coloca al final. Anade el nuevo nodo a la cache y
	 *  le pone el flag <i>dirty</i> a true.
	 *
	 *@param  key  key a buscar
	 *@param  n    nuevo nodo asignado a esta key
	 */
	private void put(Long key, AbstractNode n) {
		if (cache.containsKey(key)) {
			// remove from LRU queue and add it at the end.
			lru.remove(key);
			lru.add(key);
			CacheEntry e = (CacheEntry) cache.remove(key);
			e.dirty = true;
			e.node = n;
			cache.put(key, e);
			tree.stats.hits++;
		}
		else {
			add(key, new CacheEntry(n, true));
			tree.stats.misses++;
		}
	}


	/**
	 *  Elimina el primer objeto de la lista lru del cache.
	 *
	 *@return    Description of the Returned Value
	 */
	private CacheEntry remove() {
		// remove least recently used from both the queue and the cache.
		Object r = lru.removeFirst();
		CacheEntry e = (CacheEntry) cache.remove(r);
		if (writeBack && e.dirty) {
			super.writeNode(e.node);
		}
		return e;
	}


	/**
	 *  Anade un nuevo nodo al cache.
	 *
	 *@param  key  clave a anadir.
	 *@param  e    nueva entrada del cache.
	 */
	private void add(Long key, CacheEntry e) {
		if (lru.size() == capacity) {
			remove();
		}
		cache.put(key, e);
		lru.add(key);
	}


	/**
	 *  Description of the Class
	 *
	 *@author    Chris Seguin
	 */
	class CacheEntry {
		/**
		 *  Description of the Field
		 */
		public AbstractNode node;
		/**
		 *  Description of the Field
		 */
		public boolean dirty = false;


		/**
		 *  Constructor for the CacheEntry object
		 *
		 *@param  n  Description of Parameter
		 */
		public CacheEntry(AbstractNode n) {
			node = n;
		}


		/**
		 *  Constructor for the CacheEntry object
		 *
		 *@param  n  Description of Parameter
		 *@param  b  Description of Parameter
		 */
		public CacheEntry(AbstractNode n, boolean b) {
			node = n;
			dirty = b;
		}
	}
}
// CachedPersistentPageFile
